@@ -1,17 +1,30 @@
 /* $Id$ */
-/* Copyright (c) 2005-2015 Pierre Pronchery <khorben@defora.org> */
+/* Copyright (c) 2005-2021 Pierre Pronchery <khorben@defora.org> */
 /* This file is part of DeforaOS System libSystem */
-/* This program is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation, version 3 of the License.
+/* All rights reserved.
  *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are
+ * met:
  *
- * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>. */
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ *
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
+ * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
+ * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 
 
 
@@ -49,13 +62,14 @@ typedef struct _ConfigForeachData
 typedef struct _ConfigForeachSectionData
 {
 	ConfigForeachSectionCallback callback;
+	String const * section;
 	void * priv;
 } ConfigForeachSectionData;
 
 typedef struct _ConfigSave
 {
 	FILE * fp;
-	char const * sep;
+	String const * sep;
 } ConfigSave;
 
 
@@ -68,6 +82,74 @@ Config * config_new(void)
 }
 
 
+/* config_new_copy */
+typedef struct _ConfigError
+{
+	Config * config;
+	ErrorCode code;
+} ConfigError;
+static void _new_copy_foreach(Config const * from, String const * section,
+		void * priv);
+static void _new_copy_foreach_section(Config const * from,
+		String const * section, String const * variable,
+		String const * value, void * priv);
+
+Config * config_new_copy(Config const * from)
+{
+	ConfigError ce;
+
+	if((ce.config = mutator_new()) == NULL)
+		return NULL;
+	ce.code = 0;
+	config_foreach(from, _new_copy_foreach, &ce);
+	if(ce.code != 0)
+	{
+		config_delete(ce.config);
+		return NULL;
+	}
+	return ce.config;
+}
+
+static void _new_copy_foreach(Config const * from, String const * section,
+		void * priv)
+{
+	ConfigError * ce = (ConfigError *)priv;
+
+	if(ce->code == 0)
+		config_foreach_section(from, section, _new_copy_foreach_section,
+				ce);
+}
+
+static void _new_copy_foreach_section(Config const * from,
+		String const * section, String const * variable,
+		String const * value, void * priv)
+{
+	ConfigError * ce = (ConfigError *)priv;
+	(void) from;
+
+	if(config_set(ce->config, section, variable, value) != 0)
+		ce->code = error_get_code();
+}
+
+
+/* config_new_load */
+Config * config_new_load(String const * filename)
+{
+	Config * config;
+
+	if(filename == NULL)
+		return config_new();
+	if((config = config_new()) == NULL)
+		return NULL;
+	if(config_load(config, filename) != 0)
+	{
+		config_delete(config);
+		return NULL;
+	}
+	return config;
+}
+
+
 /* config_delete */
 void config_delete(Config * config)
 {
@@ -78,15 +160,15 @@ void config_delete(Config * config)
 
 /* accessors */
 /* config_get */
-char const * config_get(Config * config, char const * section,
-		char const * variable)
+String const * config_get(Config const * config, String const * section,
+		String const * variable)
 {
 	Mutator * mutator;
-	char const * value;
+	String const * value;
 
 	if(section == NULL)
 		section = "";
-	if((mutator = mutator_get(config, section)) == NULL)
+	if((mutator = (Mutator *)mutator_get(config, section)) == NULL)
 	{
 		/* the section does not exist */
 		if(section[0] == '\0')
@@ -95,7 +177,7 @@ char const * config_get(Config * config, char const * section,
 			error_set_code(1, "%s%s", section, ": No such section");
 		return NULL;
 	}
-	if((value = mutator_get(mutator, variable)) == NULL)
+	if((value = (String const *)mutator_get(mutator, variable)) == NULL)
 	{
 		/* the variable is not defined */
 		error_set_code(1, "%s%s%s%s%s", variable, ": Not defined in",
@@ -109,12 +191,12 @@ char const * config_get(Config * config, char const * section,
 
 
 /* config_set */
-int config_set(Config * config, char const * section, char const * variable,
-		char const * value)
+int config_set(Config * config, String const * section, String const * variable,
+		String const * value)
 {
 	Mutator * mutator;
-	char * p;
-	char * newvalue = NULL;
+	String * p;
+	String * newvalue = NULL;
 
 #ifdef DEBUG
 	fprintf(stderr, "DEBUG: %s(\"%s\", \"%s\", \"%s\")\n", __func__,
@@ -125,7 +207,7 @@ int config_set(Config * config, char const * section, char const * variable,
 	if(variable == NULL || string_get_length(variable) == 0)
 		return error_set_code(-EINVAL, "variable: %s",
 				strerror(EINVAL));
-	if((mutator = mutator_get(config, section)) == NULL)
+	if((mutator = (Mutator *)mutator_get(config, section)) == NULL)
 	{
 		/* create a new section */
 		if((mutator = mutator_new()) == NULL)
@@ -137,7 +219,8 @@ int config_set(Config * config, char const * section, char const * variable,
 		}
 		p = NULL;
 	}
-	else if((p = mutator_get(mutator, variable)) == NULL && value == NULL)
+	else if((p = (String *)mutator_get(mutator, variable)) == NULL
+			&& value == NULL)
 		/* there is nothing to do */
 		return 0;
 	if(value != NULL && (newvalue = string_new(value)) == NULL)
@@ -155,9 +238,10 @@ int config_set(Config * config, char const * section, char const * variable,
 
 /* useful */
 /* config_foreach */
-static void _foreach_callback(char const * key, void * value, void * data);
+static void _foreach_callback(Config const * config, String const * key,
+		void * value, void * data);
 
-void config_foreach(Config * config, ConfigForeachCallback callback,
+void config_foreach(Config const * config, ConfigForeachCallback callback,
 		void * priv)
 {
 	ConfigForeachData data;
@@ -167,38 +251,41 @@ void config_foreach(Config * config, ConfigForeachCallback callback,
 	mutator_foreach(config, _foreach_callback, &data);
 }
 
-static void _foreach_callback(char const * key, void * value, void * data)
+static void _foreach_callback(Config const * config, String const * key,
+		void * value, void * data)
 {
-	ConfigForeachData * priv = data;
+	ConfigForeachData * priv = (ConfigForeachData *)data;
 	(void) value;
 
-	priv->callback(key, priv->priv);
+	priv->callback(config, key, priv->priv);
 }
 
 
 /* config_foreach_section */
-static void _foreach_section_callback(char const * key, void * value,
-		void * data);
+static void _foreach_section_callback(Mutator const * mutator,
+		String const * key, void * value, void * data);
 
-void config_foreach_section(Config * config, char const * section,
+void config_foreach_section(Config const * config, String const * section,
 		ConfigForeachSectionCallback callback, void * priv)
 {
 	Mutator * mutator;
 	ConfigForeachSectionData data;
 
-	if((mutator = mutator_get(config, section)) == NULL)
+	if((mutator = (Mutator *)mutator_get(config, section)) == NULL)
 		return; /* could not find section */
 	data.callback = callback;
+	data.section = section;
 	data.priv = priv;
 	mutator_foreach(mutator, _foreach_section_callback, &data);
 }
 
-static void _foreach_section_callback(char const * key, void * value,
-		void * data)
+static void _foreach_section_callback(Mutator const * mutator,
+		String const * key, void * value, void * data)
 {
-	ConfigForeachSectionData * priv = data;
+	ConfigForeachSectionData * priv = (ConfigForeachSectionData *)data;
 
-	priv->callback(key, value, priv->priv);
+	priv->callback(mutator, priv->section, key, (String const *)value,
+			priv->priv);
 }
 
 
@@ -208,7 +295,7 @@ static String * _load_section(FILE * fp);
 static String * _load_variable(FILE * fp, int c);
 static String * _load_value(FILE * fp);
 
-int config_load(Config * config, char const * filename)
+int config_load(Config * config, String const * filename)
 {
 	int ret = 0;
 	size_t line;
@@ -270,28 +357,22 @@ static int _load_isprint(int c)
 static String * _load_section(FILE * fp)
 {
 	int c;
-	char * str = NULL;
-	size_t len = 0;
-	char * p;
+	String * str;
+	String buf[2] = "\0";
 
+	if((str = string_new("")) == NULL)
+		return NULL;
 	while((c = fgetc(fp)) != EOF && c != ']' && _load_isprint(c))
 	{
-		if((p = realloc(str, sizeof(*str) * (len + 2))) == NULL)
-		{
-			free(str);
-			return NULL;
-		}
-		str = p;
-		str[len++] = c;
+		buf[0] = c;
+		if(string_append(&str, buf) != 0)
+			break;
 	}
 	if(c != ']')
 	{
-		free(str);
+		string_delete(str);
 		return NULL;
 	}
-	if(str == NULL)
-		return string_new("");
-	str[len] = '\0';
 	return str;
 }
 
@@ -307,10 +388,7 @@ static String * _load_variable(FILE * fp, int c)
 	{
 		buf[0] = c;
 		if(string_append(&str, buf) != 0)
-		{
-			string_delete(str);
-			return NULL;
-		}
+			break;
 	}
 	if(c != '=')
 	{
@@ -324,27 +402,21 @@ static String * _load_value(FILE * fp)
 {
 	int c;
 	String * str = NULL;
-	size_t len = 0;
-	char * p;
+	String buf[2] = "\0";
 
 	while((c = fgetc(fp)) != EOF && _load_isprint(c))
 	{
-		if((p = realloc(str, sizeof(*str) * (len + 2))) == NULL)
-		{
-			free(str);
-			return NULL;
-		}
-		str = p;
-		str[len++] = c;
+		buf[0] = c;
+		if(string_append(&str, buf) != 0)
+			break;
 	}
 	if(c != EOF && c != '\n')
 	{
-		free(str);
+		string_delete(str);
 		return NULL;
 	}
 	if(str == NULL)
 		return string_new("");
-	str[len] = '\0';
 	return str;
 }
 
@@ -357,7 +429,7 @@ int config_load_preferences(Config * config, String const * vendor,
 
 	if((ret = config_load_preferences_system(config, vendor, package,
 					filename)) != 0
-			&& ret != -ENOENT)
+			&& ret != -ENOENT && ret != -EPERM)
 		return ret;
 	if((ret = config_load_preferences_user(config, vendor, package,
 					filename)) != 0
@@ -376,9 +448,11 @@ int config_load_preferences_system(Config * config, String const * vendor,
 
 	if(filename == NULL)
 		return error_set_code(-EINVAL, "%s", strerror(EINVAL));
-	if(vendor != NULL && string_find(vendor, "/") != NULL)
+	if(vendor != NULL && (string_compare_length(vendor, "../", 3) == 0
+				|| string_find(vendor, "/..") != NULL))
 		return error_set_code(-EPERM, "%s", strerror(EPERM));
-	if(package != NULL && string_find(package, "/") != NULL)
+	if(package != NULL && (string_compare_length(package, "../", 3) == 0
+				|| string_find(package, "/..") != NULL))
 		return error_set_code(-EPERM, "%s", strerror(EPERM));
 	if((f = string_new_append(SYSCONFDIR, "/",
 					(vendor != NULL) ? vendor : "", "/",
@@ -401,9 +475,11 @@ int config_load_preferences_user(Config * config, String const * vendor,
 
 	if(filename == NULL)
 		return error_set_code(-EINVAL, "%s", strerror(EINVAL));
-	if(vendor != NULL && string_find(vendor, "/") != NULL)
+	if(vendor != NULL && (string_compare_length(vendor, "../", 3) == 0
+				|| string_find(vendor, "/..") != NULL))
 		return error_set_code(-EPERM, "%s", strerror(EPERM));
-	if(package != NULL && string_find(package, "/") != NULL)
+	if(package != NULL && (string_compare_length(package, "../", 3) == 0
+				|| string_find(package, "/..") != NULL))
 		return error_set_code(-EPERM, "%s", strerror(EPERM));
 	if(filename != NULL && string_find(filename, "/") != NULL)
 		return error_set_code(-EPERM, "%s", strerror(EPERM));
@@ -421,9 +497,10 @@ int config_load_preferences_user(Config * config, String const * vendor,
 
 
 /* config_reset */
-static void _delete_foreach(String const * key, void * value, void * data);
-static void _delete_foreach_section(String const * key, void * value,
-		void * data);
+static void _delete_foreach(Mutator const * mutator, String const * key,
+		void * value, void * data);
+static void _delete_foreach_section(Mutator const * mutator, String const * key,
+		void * value, void * data);
 
 int config_reset(Config * config)
 {
@@ -431,21 +508,24 @@ int config_reset(Config * config)
 	return mutator_reset(config);
 }
 
-static void _delete_foreach(String const * key, void * value, void * data)
+static void _delete_foreach(Mutator const * mutator, String const * key,
+		void * value, void * data)
 {
-	Mutator * mutator = value;
+	Mutator * m = (Mutator *)value;
+	(void) mutator;
 	(void) key;
 	(void) data;
 
 	/* free the values */
-	mutator_foreach(mutator, _delete_foreach_section, NULL);
-	mutator_delete(mutator);
+	mutator_foreach(m, _delete_foreach_section, NULL);
+	mutator_delete(m);
 }
 
-static void _delete_foreach_section(String const * key, void * value,
-		void * data)
+static void _delete_foreach_section(Mutator const * mutator, String const * key,
+		void * value, void * data)
 {
-	String * v = value;
+	String * v = (String *)value;
+	(void) mutator;
 	(void) key;
 	(void) data;
 
@@ -454,12 +534,14 @@ static void _delete_foreach_section(String const * key, void * value,
 
 
 /* config_save */
-static void _save_foreach_default(char const * section, void * value,
-		void * data);
-static void _save_foreach(char const * section, void * value, void * data);
-static void _save_foreach_section(char const * key, void * value, void * data);
+static void _save_foreach_default(Mutator const * mutator,
+		String const * section, void * value, void * data);
+static void _save_foreach(Mutator const * mutator, String const * section,
+		void * value, void * data);
+static void _save_foreach_section(Mutator const * mutator, String const * key,
+		void * value, void * data);
 
-int config_save(Config * config, char const * filename)
+int config_save(Config const * config, String const * filename)
 {
 	ConfigSave save;
 
@@ -481,23 +563,26 @@ int config_save(Config * config, char const * filename)
 	return 0;
 }
 
-static void _save_foreach_default(char const * section, void * value,
-		void * data)
+static void _save_foreach_default(Mutator const * mutator,
+		String const * section, void * value, void * data)
 {
-	ConfigSave * save = data;
-	Mutator * mutator = value;
+	ConfigSave * save = (ConfigSave *)data;
+	Mutator * m = (Mutator *)value;
+	(void) mutator;
 
 	if(save->fp == NULL)
 		return;
 	if(section[0] != '\0')
 		return;
-	mutator_foreach(mutator, _save_foreach_section, save);
+	mutator_foreach(m, _save_foreach_section, save);
 }
 
-static void _save_foreach(char const * section, void * value, void * data)
+static void _save_foreach(Mutator const * mutator, String const * section,
+		void * value, void * data)
 {
-	ConfigSave * save = data;
-	Mutator * mutator = value;
+	ConfigSave * save = (ConfigSave *)data;
+	Mutator * m = (Mutator *)value;
+	(void) mutator;
 
 	if(save->fp == NULL)
 		return;
@@ -510,17 +595,18 @@ static void _save_foreach(char const * section, void * value, void * data)
 		return;
 	}
 	save->sep = "\n";
-	mutator_foreach(mutator, _save_foreach_section, save);
+	mutator_foreach(m, _save_foreach_section, save);
 }
 
-static void _save_foreach_section(char const * key, void * value, void * data)
+static void _save_foreach_section(Mutator const * mutator, String const * key,
+		void * value, void * data)
 {
-	ConfigSave * save = data;
-	char const * val = value;
+	ConfigSave * save = (ConfigSave *)data;
+	String const * val = (String const *)value;
+	(void) mutator;
 
 	if(save->fp == NULL)
 		return;
-	/* FIXME escape lines with a backslash */
 	if(val == NULL)
 		return;
 	if(fprintf(save->fp, "%s%s=%s", save->sep, key, val) < 0)
@@ -534,9 +620,9 @@ static void _save_foreach_section(char const * key, void * value, void * data)
 
 
 /* config_save_preferences_user */
-static int _save_preferences_user_append(String ** f, String const * dir);
+static int _save_preferences_user_mkdir(String * dir);
 
-int config_save_preferences_user(Config * config, String const * vendor,
+int config_save_preferences_user(Config const * config, String const * vendor,
 		String const * package, String const * filename)
 {
 	int ret;
@@ -545,37 +631,54 @@ int config_save_preferences_user(Config * config, String const * vendor,
 
 	if(filename == NULL)
 		return error_set_code(-EINVAL, "%s", strerror(EINVAL));
-	if(vendor != NULL && string_find(vendor, "/") != NULL)
-		return error_set_code(-EPERM, "%s", strerror(EPERM));
-	if(package != NULL && string_find(package, "/") != NULL)
-		return error_set_code(-EPERM, "%s", strerror(EPERM));
-	if(filename != NULL && string_find(filename, "/") != NULL)
-		return error_set_code(-EPERM, "%s", strerror(EPERM));
 	if((homedir = getenv("HOME")) == NULL)
 		return error_set_code(-errno, "%s", strerror(errno));
-	if((f = string_new_append(homedir, "/.config", NULL)) == NULL)
+	if((f = string_new_append(homedir, "/.config/",
+					(vendor != NULL) ? vendor : "",
+					(vendor != NULL) ? "/" : "",
+					(package != NULL) ? package : "",
+					(package != NULL) ? "/" : "",
+					filename, NULL)) == NULL)
 		return error_get_code();
-	if(vendor != NULL && (ret = _save_preferences_user_append(&f,
-					vendor)) != 0)
-		return ret;
-	if(package != NULL && (ret = _save_preferences_user_append(&f,
-					package)) != 0)
-		return ret;
-	if(string_append(&f, "/") != 0 || string_append(&f, filename) != 0)
-		return error_get_code();
-	ret = config_save(config, f);
+	if((ret = _save_preferences_user_mkdir(f)) == 0)
+		ret = config_save(config, f);
 	string_delete(f);
 	return ret;
 }
 
-static int _save_preferences_user_append(String ** f, String const * dir)
+static int _save_preferences_user_mkdir(String * dir)
 {
-	if(string_append(f, "/") != 0)
-		return error_get_code();
-	if(string_append(f, dir) != 0)
-		return error_get_code();
-	if(mkdir(*f, 0755) != 0
-			&& errno != EEXIST)
-		return error_set_code(-errno, "%s: %s", *f, strerror(errno));
+	int ret = 0;
+	size_t i;
+	size_t j;
+	size_t len;
+
+#ifdef DEBUG
+	fprintf(stderr, "DEBUG: %s(\"%s\")\n", __func__, dir);
+#endif
+	len = string_get_length(dir);
+	for(i = 0, j = 0; j < len; j++)
+	{
+		if(dir[j] != '/')
+			continue;
+		if(i == j)
+		{
+			i++;
+			continue;
+		}
+		dir[j] = '\0';
+#ifdef DEBUG
+		fprintf(stderr, "DEBUG: %s() \"%s\"\n", __func__, dir);
+#endif
+		if(dir[i] == '.')
+			ret = error_set_code(-EPERM, "%s: %s", dir,
+					strerror(EPERM));
+		else if(mkdir(dir, 0777) != 0 && errno != EEXIST)
+			ret = error_set_code(-errno, "%s: %s", dir,
+					strerror(errno));
+		dir[j] = '/';
+		if(ret != 0)
+			return ret;
+	}
 	return 0;
 }
