@@ -1,17 +1,30 @@
 /* $Id$ */
-/* Copyright (c) 2005-2015 Pierre Pronchery <khorben@defora.org> */
+/* Copyright (c) 2005-2020 Pierre Pronchery <khorben@defora.org> */
 /* This file is part of DeforaOS System libSystem */
-/* This program is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation, version 3 of the License.
+/* All rights reserved.
  *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are
+ * met:
  *
- * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>. */
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ *
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
+ * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
+ * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 
 
 
@@ -53,7 +66,7 @@ typedef struct _EventTimeout
 	EventTimeoutFunc func;
 	void * data;
 } EventTimeout;
-ARRAY(EventTimeout *, eventtimeout)
+ARRAY2(EventTimeout *, eventtimeout)
 
 typedef struct _EventIO
 {
@@ -61,7 +74,7 @@ typedef struct _EventIO
 	EventIOFunc func;
 	void * data;
 } EventIO;
-ARRAY(EventIO *, eventio)
+ARRAY2(EventIO *, eventio)
 
 struct _Event
 {
@@ -76,6 +89,10 @@ struct _Event
 };
 
 
+/* prototypes */
+static int _event_loop_once(Event * event);
+
+
 /* public */
 /* functions */
 /* event_new */
@@ -83,7 +100,7 @@ Event * event_new(void)
 {
 	Event * event;
 
-	if((event = object_new(sizeof(*event))) == NULL)
+	if((event = (Event *)object_new(sizeof(*event))) == NULL)
 		return NULL;
 	event->timeouts = eventtimeoutarray_new();
 	event->loop = 0;
@@ -135,133 +152,13 @@ void event_delete(Event * event)
 
 /* useful */
 /* event_loop */
-static int _loop_timeout(Event * event);
-static void _loop_io(Event * event, eventioArray * eios, fd_set * fds);
-
 int event_loop(Event * event)
 {
-	struct timeval tv = event->timeout;
-	struct timeval * timeout = (tv.tv_sec == (time_t)LONG_MAX
-			&& tv.tv_usec == (suseconds_t)LONG_MAX) ? NULL : &tv;
-	fd_set rfds = event->rfds;
-	fd_set wfds = event->wfds;
+	int ret;
 
 	event->loop++;
-	while(event->loop && (timeout != NULL || event->fdmax != -1))
-	{
-		if(select(event->fdmax + 1, &rfds, &wfds, NULL, timeout) < 0)
-			return -error_set_code(1, "%s", strerror(errno));
-		if(_loop_timeout(event) != 0)
-			return -1;
-		_loop_io(event, event->reads, &rfds);
-		_loop_io(event, event->writes, &wfds);
-		if(event->timeout.tv_sec == (time_t)LONG_MAX
-				&& event->timeout.tv_usec
-				== (suseconds_t)LONG_MAX)
-			timeout = NULL;
-		else
-			timeout = &event->timeout;
-		rfds = event->rfds;
-		wfds = event->wfds;
-	}
-	return 0;
-}
-
-static int _loop_timeout(Event * event)
-{
-	struct timeval now;
-	unsigned int i = 0;
-	EventTimeout * et;
-
-	if(gettimeofday(&now, NULL) != 0)
-	{
-		error_set_code(1, "%s", strerror(errno));
-		return -1;
-	}
-	event->timeout.tv_sec = (time_t)LONG_MAX;
-	event->timeout.tv_usec = (suseconds_t)LONG_MAX;
-	while(i < array_count(event->timeouts))
-	{
-		array_get_copy(event->timeouts, i, &et);
-		if(now.tv_sec > et->timeout.tv_sec
-				|| (now.tv_sec == et->timeout.tv_sec
-					&& now.tv_usec >= et->timeout.tv_usec))
-		{
-			if(et->func(et->data) != 0)
-			{
-				array_remove_pos(event->timeouts, i);
-				object_delete(et);
-				continue;
-			}
-			et->timeout.tv_sec = et->initial.tv_sec + now.tv_sec;
-			et->timeout.tv_usec = et->initial.tv_usec + now.tv_usec;
-			if(et->initial.tv_sec < event->timeout.tv_sec
-					|| (et->initial.tv_sec
-						== event->timeout.tv_sec
-						&& et->initial.tv_usec
-						< event->timeout.tv_usec))
-			{
-				event->timeout.tv_sec = et->initial.tv_sec;
-				event->timeout.tv_usec = et->initial.tv_usec;
-			}
-		}
-		else
-		{
-			if(et->timeout.tv_sec - now.tv_sec < event->timeout.tv_sec
-					|| (et->timeout.tv_sec - now.tv_sec == event->timeout.tv_sec
-						&& et->timeout.tv_usec - now.tv_usec < event->timeout.tv_usec))
-			{
-				event->timeout.tv_sec = et->timeout.tv_sec
-					- now.tv_sec;
-				/* FIXME may be needed elsewhere too */
-				if(et->timeout.tv_usec >= now.tv_usec)
-					event->timeout.tv_usec
-						= et->timeout.tv_usec
-						- now.tv_usec;
-				else
-				{
-					event->timeout.tv_sec--;
-					event->timeout.tv_usec
-						= now.tv_usec
-						- et->timeout.tv_usec;
-				}
-			}
-		}
-		i++;
-	}
-#ifdef DEBUG
-	fprintf(stderr, "DEBUG: %s() %s%ld%s%ld => 0\n", __func__, "tv_sec=",
-			(long)event->timeout.tv_sec, ", tv_usec=",
-			(long)event->timeout.tv_usec);
-#endif
-	return 0;
-}
-
-static void _loop_io(Event * event, eventioArray * eios, fd_set * fds)
-{
-	unsigned int i = 0;
-	EventIO * eio;
-	int fd;
-
-	while(i < array_count(eios))
-	{
-		array_get_copy(eios, i, &eio);
-		if((fd = eio->fd) <= event->fdmax && FD_ISSET(fd, fds)
-				&& eio->func(fd, eio->data) != 0)
-		{
-			if(eios == event->reads)
-				event_unregister_io_read(event, fd);
-			else if(eios == event->writes)
-				event_unregister_io_write(event, fd);
-#ifdef DEBUG
-			else
-				fprintf(stderr, "DEBUG: %s%s", __func__,
-						"(): should not happen\n");
-#endif
-		}
-		else
-			i++;
-	}
+	while(event->loop && (ret = _event_loop_once(event)) == 0);
+	return ret;
 }
 
 
@@ -273,6 +170,19 @@ void event_loop_quit(Event * event)
 }
 
 
+/* event_loop_while */
+int event_loop_while(Event * event, const int * flag)
+{
+	int ret;
+
+	if(flag == NULL)
+		return event_loop(event);
+	event->loop++;
+	while(event->loop && *flag && (ret = _event_loop_once(event)) == 0);
+	return ret;
+}
+
+
 /* event_register_idle */
 int event_register_idle(Event * event, EventTimeoutFunc func, void * data)
 {
@@ -291,7 +201,7 @@ int event_register_io_read(Event * event, int fd, EventIOFunc func,
 	EventIO * eventio;
 
 	assert(fd >= 0);
-	if((eventio = object_new(sizeof(*eventio))) == NULL)
+	if((eventio = (EventIO *)object_new(sizeof(*eventio))) == NULL)
 		return 1;
 	eventio->fd = fd;
 	eventio->func = func;
@@ -314,7 +224,7 @@ int event_register_io_write(Event * event, int fd, EventIOFunc func,
 	EventIO * eventio;
 
 	assert(fd >= 0);
-	if((eventio = object_new(sizeof(*eventio))) == NULL)
+	if((eventio = (EventIO *)object_new(sizeof(*eventio))) == NULL)
 		return 1;
 	eventio->fd = fd;
 	eventio->func = func;
@@ -338,8 +248,9 @@ int event_register_timeout(Event * event, struct timeval * timeout,
 	struct timeval now;
 
 	if(gettimeofday(&now, NULL) != 0)
-		return -error_set_code(1, "%s", strerror(errno));
-	if((eventtimeout = object_new(sizeof(*eventtimeout))) == NULL)
+		return error_set_code(-errno, "%s", strerror(errno));
+	if((eventtimeout = (EventTimeout *)object_new(sizeof(*eventtimeout)))
+			== NULL)
 		return -1;
 	eventtimeout->initial.tv_sec = timeout->tv_sec;
 	eventtimeout->initial.tv_usec = timeout->tv_usec;
@@ -357,8 +268,8 @@ int event_register_timeout(Event * event, struct timeval * timeout,
 				&& event->timeout.tv_usec > timeout->tv_usec))
 	{
 #ifdef DEBUG
-		fprintf(stderr, "DEBUG: %s%s%ld%s%ld%s", __func__, "() tv_sec=",
-				(long)timeout->tv_sec, ", tv_usec=",
+		fprintf(stderr, "DEBUG: %s%s%lld%s%ld%s", __func__, "() tv_sec=",
+				(long long)timeout->tv_sec, ", tv_usec=",
 				(long)timeout->tv_usec, "\n");
 #endif
 		event->timeout.tv_sec = timeout->tv_sec;
@@ -431,7 +342,8 @@ int event_unregister_timeout(Event * event, EventTimeoutFunc func)
 		object_delete(et);
 	}
 	if(gettimeofday(&now, NULL) != 0)
-		return error_set_code(1, "%s", strerror(errno));
+		return error_set_code(-errno, "%s", strerror(errno));
+	/* XXX will fail in 2038 on 32-bit platforms */
 	event->timeout.tv_sec = (time_t)LONG_MAX;
 	event->timeout.tv_usec = (suseconds_t)LONG_MAX;
 	for(i = 0; i < array_count(event->timeouts); i++)
@@ -459,3 +371,130 @@ int event_unregister_timeout(Event * event, EventTimeoutFunc func)
 	}
 	return 0;
 }
+
+
+/* private */
+/* functions */
+/* event_loop_once */
+static int _loop_timeout(Event * event);
+static void _loop_io(Event * event, eventioArray * eios, fd_set * fds);
+
+static int _event_loop_once(Event * event)
+{
+	struct timeval tv = event->timeout;
+	struct timeval * timeout = (tv.tv_sec == (time_t)LONG_MAX
+			&& tv.tv_usec == (suseconds_t)LONG_MAX) ? NULL : &tv;
+	fd_set rfds = event->rfds;
+	fd_set wfds = event->wfds;
+
+	if(timeout != NULL || event->fdmax != -1)
+	{
+		if(select(event->fdmax + 1, &rfds, &wfds, NULL, timeout) < 0)
+			return error_set_code(-errno, "%s", strerror(errno));
+		if(_loop_timeout(event) != 0)
+			return -1;
+		_loop_io(event, event->reads, &rfds);
+		_loop_io(event, event->writes, &wfds);
+		if(event->timeout.tv_sec == (time_t)LONG_MAX
+				&& event->timeout.tv_usec
+				== (suseconds_t)LONG_MAX)
+			timeout = NULL;
+		else
+			timeout = &event->timeout;
+	}
+	return 0;
+}
+
+static int _loop_timeout(Event * event)
+{
+	struct timeval now;
+	unsigned int i = 0;
+	EventTimeout * et;
+
+	if(gettimeofday(&now, NULL) != 0)
+		return error_set_code(-errno, "%s", strerror(errno));
+	event->timeout.tv_sec = (time_t)LONG_MAX;
+	event->timeout.tv_usec = (suseconds_t)LONG_MAX;
+	while(i < array_count(event->timeouts))
+	{
+		array_get_copy(event->timeouts, i, &et);
+		if(now.tv_sec > et->timeout.tv_sec
+				|| (now.tv_sec == et->timeout.tv_sec
+					&& now.tv_usec >= et->timeout.tv_usec))
+		{
+			if(et->func(et->data) != 0)
+			{
+				array_remove_pos(event->timeouts, i);
+				object_delete(et);
+				continue;
+			}
+			et->timeout.tv_sec = et->initial.tv_sec + now.tv_sec;
+			et->timeout.tv_usec = et->initial.tv_usec + now.tv_usec;
+			if(et->initial.tv_sec < event->timeout.tv_sec
+					|| (et->initial.tv_sec
+						== event->timeout.tv_sec
+						&& et->initial.tv_usec
+						< event->timeout.tv_usec))
+			{
+				event->timeout.tv_sec = et->initial.tv_sec;
+				event->timeout.tv_usec = et->initial.tv_usec;
+			}
+		}
+		else
+		{
+			if(et->timeout.tv_sec - now.tv_sec < event->timeout.tv_sec
+					|| (et->timeout.tv_sec - now.tv_sec == event->timeout.tv_sec
+						&& et->timeout.tv_usec - now.tv_usec < event->timeout.tv_usec))
+			{
+				event->timeout.tv_sec = et->timeout.tv_sec
+					- now.tv_sec;
+				/* XXX may be needed elsewhere too */
+				if(et->timeout.tv_usec >= now.tv_usec)
+					event->timeout.tv_usec
+						= et->timeout.tv_usec
+						- now.tv_usec;
+				else
+				{
+					event->timeout.tv_sec--;
+					event->timeout.tv_usec
+						= now.tv_usec
+						- et->timeout.tv_usec;
+				}
+			}
+		}
+		i++;
+	}
+#ifdef DEBUG
+	fprintf(stderr, "DEBUG: %s() %s%lld%s%ld => 0\n", __func__, "tv_sec=",
+			(long long)event->timeout.tv_sec, ", tv_usec=",
+			(long)event->timeout.tv_usec);
+#endif
+	return 0;
+}
+
+static void _loop_io(Event * event, eventioArray * eios, fd_set * fds)
+{
+	unsigned int i = 0;
+	EventIO * eio;
+	int fd;
+
+	while(i < array_count(eios))
+	{
+		array_get_copy(eios, i, &eio);
+		if((fd = eio->fd) <= event->fdmax && FD_ISSET(fd, fds)
+				&& eio->func(fd, eio->data) != 0)
+		{
+			if(eios == event->reads)
+				event_unregister_io_read(event, fd);
+			else if(eios == event->writes)
+				event_unregister_io_write(event, fd);
+#ifdef DEBUG
+			else
+				fprintf(stderr, "DEBUG: %s%s", __func__,
+						"(): should not happen\n");
+#endif
+		}
+		else
+			i++;
+	}
+}
