@@ -1,17 +1,30 @@
 /* $Id$ */
-/* Copyright (c) 2005-2015 Pierre Pronchery <khorben@defora.org> */
+/* Copyright (c) 2005-2021 Pierre Pronchery <khorben@defora.org> */
 /* This file is part of DeforaOS System libSystem */
-/* This program is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation, version 3 of the License.
+/* All rights reserved.
  *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are
+ * met:
  *
- * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>. */
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ *
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
+ * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
+ * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 
 
 
@@ -32,7 +45,7 @@ typedef struct _HashEntry
 	void const * key;
 	void * value;
 } HashEntry;
-ARRAY(HashEntry, _hashentry)
+ARRAY2(HashEntry, _hashentry)
 #define HashEntryArray _hashentryArray
 
 
@@ -87,7 +100,7 @@ Hash * hash_new(HashFunc func, HashCompare compare)
 		error_set_code(1, "%s", "Invalid comparison function");
 		return NULL;
 	}
-	if((hash = object_new(sizeof(*hash))) == NULL)
+	if((hash = (Hash *)object_new(sizeof(*hash))) == NULL)
 		return NULL;
 	if((hash->entries = _hashentryarray_new()) == NULL)
 	{
@@ -100,6 +113,24 @@ Hash * hash_new(HashFunc func, HashCompare compare)
 }
 
 
+/* hash_new_copy */
+Hash * hash_new_copy(Hash const * from)
+{
+	Hash * hash;
+
+	if((hash = (Hash *)object_new(sizeof(*from))) == NULL)
+		return NULL;
+	if((hash->entries = array_new_copy(from->entries)) == NULL)
+	{
+		object_delete(hash);
+		return NULL;
+	}
+	hash->func = from->func;
+	hash->compare = from->compare;
+	return hash;
+}
+
+
 /* hash_delete */
 void hash_delete(Hash * hash)
 {
@@ -112,7 +143,7 @@ void hash_delete(Hash * hash)
 /* hash_func_string */
 unsigned int hash_func_string(void const * key)
 {
-	String const * str = key;
+	String const * str = (String const *)key;
 	size_t i;
 	unsigned int hash = 0;
 
@@ -125,8 +156,8 @@ unsigned int hash_func_string(void const * key)
 /* hash_compare_string */
 int hash_compare_string(void const * value1, void const * value2)
 {
-	String const * str1 = value1;
-	String const * str2 = value2;
+	String const * str1 = (String const *)value1;
+	String const * str2 = (String const *)value2;
 
 	return string_compare(str1, str2);
 }
@@ -134,23 +165,24 @@ int hash_compare_string(void const * value1, void const * value2)
 
 /* accessors */
 /* hash_count */
-size_t hash_count(Hash * hash)
+size_t hash_count(Hash const * hash)
 {
-	return array_count(hash->entries);
+	return array_count((Array const *)hash->entries);
 }
 
 
 /* hash_get */
-void * hash_get(Hash * hash, void const * key)
+void * hash_get(Hash const * hash, void const * key)
 {
+	Array const * entries = (Array const *)hash->entries;
 	unsigned int h;
 	size_t i;
 	HashEntry * he;
 
 	h = (hash->func != NULL) ? hash->func(key) : 0;
-	for(i = array_count(hash->entries); i > 0; i--)
+	for(i = array_count(entries); i > 0; i--)
 	{
-		if((he = array_get(hash->entries, i - 1)) == NULL)
+		if((he = (HashEntry *)array_get(entries, i - 1)) == NULL)
 			return NULL;
 		if(he->hash != h)
 			continue;
@@ -163,16 +195,17 @@ void * hash_get(Hash * hash, void const * key)
 
 
 /* hash_get_key */
-void * hash_get_key(Hash * hash, void const * key)
+void const * hash_get_key(Hash const * hash, void const * key)
 {
+	Array const * entries = (Array const *)hash->entries;
 	unsigned int h;
 	size_t i;
-	HashEntry * he;
+	HashEntry const * he;
 
 	h = (hash->func != NULL) ? hash->func(key) : 0;
-	for(i = array_count(hash->entries); i > 0; i--)
+	for(i = array_count(entries); i > 0; i--)
 	{
-		if((he = array_get(hash->entries, i - 1)) == NULL)
+		if((he = (HashEntry const *)array_get(entries, i - 1)) == NULL)
 			return NULL;
 		if(he->hash != h)
 			continue;
@@ -187,6 +220,7 @@ void * hash_get_key(Hash * hash, void const * key)
 /* hash_set */
 int hash_set(Hash * hash, void const * key, void * value)
 {
+	Array * entries = (Array *)hash->entries;
 	unsigned int h;
 	size_t i;
 	size_t cnt;
@@ -194,23 +228,22 @@ int hash_set(Hash * hash, void const * key, void * value)
 	HashEntry * p;
 
 	h = (hash->func != NULL) ? hash->func(key) : 0;
-	for(i = 0, cnt = array_count(hash->entries); i < cnt; i++)
+	for(i = 0, cnt = array_count(entries); i < cnt; i++)
 	{
-		if((p = array_get(hash->entries, i)) == NULL)
+		if((p = (HashEntry *)array_get(entries, i)) == NULL)
 			return 1;
 		if(p->hash != h)
 			continue;
 		if(hash->compare(p->key, key) != 0)
 			continue;
 		if(value == NULL)
-			return (array_remove_pos(hash->entries, i) == 0)
-				? 0 : 1;
+			return (array_remove_pos(entries, i) == 0) ? 0 : 1;
 		return _hashentry_set_value(p, value);
 	}
 	if(value == NULL)
 		return 0;
 	_hashentry_init(&he, hash->func, key, value);
-	return (array_append(hash->entries, &he) == 0) ? 0 : 1;
+	return (array_append(entries, &he) == 0) ? 0 : 1;
 }
 
 
@@ -220,33 +253,38 @@ static void _hash_foreach(void * value, void * data);
 
 struct funcdata
 {
+	Hash const * hash;
 	HashForeach func;
 	void * data;
 };
 
-void hash_foreach(Hash * hash, HashForeach func, void * data)
+void hash_foreach(Hash const * hash, HashForeach func, void * data)
 {
+	Array const * entries = (Array const *)hash->entries;
 	struct funcdata fd;
 
+	fd.hash = hash;
 	fd.func = func;
 	fd.data = data;
-	array_foreach(hash->entries, _hash_foreach, &fd);
+	array_foreach(entries, _hash_foreach, &fd);
 }
 
 static void _hash_foreach(void * value, void * data)
 {
-	HashEntry * he = value;
-	struct funcdata * fd = data;
+	HashEntry * he = (HashEntry *)value;
+	struct funcdata * fd = (struct funcdata *)data;
 
-	fd->func(he->key, he->value, fd->data);
+	fd->func(fd->hash, he->key, he->value, fd->data);
 }
 
 
 /* hash_reset */
 int hash_reset(Hash * hash)
 {
-	while(array_count(hash->entries))
-		if(array_remove_pos(hash->entries, 0) != 0)
+	Array * entries = (Array *)hash->entries;
+
+	while(array_count(entries))
+		if(array_remove_pos(entries, 0) != 0)
 			return 1;
 	return 0;
 }
