@@ -1,17 +1,30 @@
 /* $Id$ */
-/* Copyright (c) 2008-2015 Pierre Pronchery <khorben@defora.org> */
+/* Copyright (c) 2008-2020 Pierre Pronchery <khorben@defora.org> */
 /* This file is part of DeforaOS System libSystem */
-/* This program is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation, version 3 of the License.
+/* All rights reserved.
  *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are
+ * met:
  *
- * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>. */
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ *
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
+ * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
+ * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 /* TODO:
  * - switch parser_scan() and parser_scan_filter() to avoid confusion */
 
@@ -39,11 +52,12 @@ struct _Parser
 	/* parsing sources */
 	String * filename;
 	FILE * fp;
-	char * string;
+	String * string;
 	size_t string_cnt;
 	size_t string_pos;
 
 	/* tracking the position */
+	int error;
 	unsigned int line;
 	unsigned int col;
 	int last;
@@ -90,12 +104,18 @@ int parser_scan_filter(Parser * parser)
 	if(parser->lookahead)
 		parser->lookahead--;
 	else if(parser->scanner(&c, parser) != 0)
-		return EOF; /* FIXME report error */
+	{
+		parser->error = 1;
+		return EOF;
+	}
 	for(i = 0; i < parser->filters_cnt; i++)
 	{
 		pfd = &parser->filters[i];
 		if((l = pfd->filter(&c, pfd->data)) < 0)
+		{
+			parser->error = 1;
 			return EOF;
+		}
 		parser->lookahead += l;
 	}
 	parser->last = c;
@@ -106,7 +126,7 @@ int parser_scan_filter(Parser * parser)
 /* parser_scanner_file */
 static int _parser_scanner_file(int * c, void * data)
 {
-	Parser * parser = data;
+	Parser * parser = (Parser *)data;
 
 #ifdef DEBUG
 	fprintf(stderr, "DEBUG: %s()\n", __func__);
@@ -129,7 +149,7 @@ static int _parser_scanner_file(int * c, void * data)
 /* parser_scanner_string */
 static int _parser_scanner_string(int * c, void * data)
 {
-	Parser * parser = data;
+	Parser * parser = (Parser *)data;
 
 #ifdef DEBUG
 	fprintf(stderr, "DEBUG: %s()\n", __func__);
@@ -156,7 +176,7 @@ static int _parser_scanner_string(int * c, void * data)
 /* parser_new */
 static Parser * _new_do(ParserFilter scanner);
 
-Parser * parser_new(char const * pathname)
+Parser * parser_new(String const * pathname)
 {
 	Parser * parser;
 
@@ -166,9 +186,9 @@ Parser * parser_new(char const * pathname)
 	if((parser = _new_do(_parser_scanner_file)) == NULL)
 		return NULL;
 	if((parser->filename = string_new(pathname)) == NULL)
-		error_set_code(1, "%s", strerror(errno));
+		error_set_code(-errno, "%s", strerror(errno));
 	if((parser->fp = fopen(pathname, "r")) == NULL)
-		error_set_code(1, "%s: %s", pathname, strerror(errno));
+		error_set_code(-errno, "%s: %s", pathname, strerror(errno));
 	if(parser->filename == NULL || parser->fp == NULL)
 	{
 		parser_delete(parser);
@@ -181,13 +201,14 @@ static Parser * _new_do(ParserFilter scanner)
 {
 	Parser * parser;
 
-	if((parser = object_new(sizeof(*parser))) == NULL)
+	if((parser = (Parser *)object_new(sizeof(*parser))) == NULL)
 		return NULL;
 	parser->filename = NULL;
 	parser->fp = NULL;
 	parser->string = NULL;
 	parser->string_cnt = 0;
 	parser->string_pos = 0;
+	parser->error = 0;
 	parser->line = 1;
 	parser->col = 1;
 	parser->last = EOF;
@@ -202,7 +223,7 @@ static Parser * _new_do(ParserFilter scanner)
 
 
 /* parser_new_string */
-Parser * parser_new_string(char const * string, size_t length)
+Parser * parser_new_string(String const * string, size_t length)
 {
 	Parser * parser;
 
@@ -211,11 +232,11 @@ Parser * parser_new_string(char const * string, size_t length)
 #endif
 	if((parser = _new_do(_parser_scanner_string)) == NULL)
 		return NULL;
-	parser->string = malloc(length);
+	parser->string = (String *)malloc(length);
 	parser->string_cnt = length;
 	if(parser->string == NULL && length != 0)
 	{
-		error_set_code(1, "%s", strerror(errno));
+		error_set_code(-errno, "%s", strerror(errno));
 		parser_delete(parser);
 		return NULL;
 	}
@@ -235,7 +256,7 @@ int parser_delete(Parser * parser)
 #endif
 	if(parser->fp != NULL
 			&& fclose(parser->fp) != 0)
-		ret = error_set_code(1, "%s: %s", parser->filename,
+		ret = error_set_code(-errno, "%s: %s", parser->filename,
 				strerror(errno));
 	string_delete(parser->filename);
 	free(parser->string);
@@ -248,7 +269,7 @@ int parser_delete(Parser * parser)
 
 /* accessors */
 /* parser_get_filename */
-char const * parser_get_filename(Parser * parser)
+String const * parser_get_filename(Parser * parser)
 {
 	return parser->filename;
 }
@@ -263,9 +284,15 @@ int parser_get_token(Parser * parser, Token ** token)
 
 	if((*token = token_new(parser->filename, parser->line, parser->col))
 			== NULL)
-		return 1;
-	if(parser->last == EOF)
-		parser_scan_filter(parser);
+		return -1;
+	if(parser->last == EOF
+			&& parser_scan_filter(parser) == EOF
+			&& parser->error != 0)
+	{
+		token_delete(*token);
+		*token = NULL;
+		return -1;
+	}
 #ifdef DEBUG
 	fprintf(stderr, "DEBUG: %s() %c\n", __func__, parser->last);
 #endif
@@ -280,7 +307,7 @@ int parser_get_token(Parser * parser, Token ** token)
 		return 0; /* there is a token and no error */
 	token_delete(*token);
 	*token = NULL;
-	return (ret >= 0 && parser->last == EOF) ? 0 : 1;
+	return (ret >= 0 && parser->last == EOF) ? 0 : -1;
 }
 
 
@@ -290,7 +317,8 @@ int parser_add_callback(Parser * parser, ParserCallback callback, void * data)
 {
 	ParserCallbackData * p;
 
-	if((p = realloc(parser->callbacks, sizeof(*p) * (parser->callbacks_cnt
+	if((p = (ParserCallbackData *)realloc(parser->callbacks,
+					sizeof(*p) * (parser->callbacks_cnt
 						+ 1))) == NULL)
 		return 1;
 	parser->callbacks = p;
@@ -310,7 +338,8 @@ int parser_add_filter(Parser * parser, ParserFilter filter, void * data)
 	fprintf(stderr, "DEBUG: %s(%p, %p, %p)\n", __func__, (void *)parser,
 			(void *)filter, (void *)data);
 #endif
-	if((p = realloc(p, sizeof(*p) * (parser->filters_cnt + 1))) == NULL)
+	if((p = (ParserFilterData *)realloc(p, sizeof(*p) * (parser->filters_cnt
+						+ 1))) == NULL)
 		return 1;
 	parser->filters = p;
 	p = &parser->filters[parser->filters_cnt++];
@@ -322,7 +351,7 @@ int parser_add_filter(Parser * parser, ParserFilter filter, void * data)
 
 /* parser_remove_callback */
 int parser_remove_callback(Parser * parser, ParserCallback callback)
-	/* FIXME untested */
+	/* XXX untested */
 {
 	size_t i;
 	ParserCallbackData * p;
@@ -346,7 +375,7 @@ int parser_remove_callback(Parser * parser, ParserCallback callback)
 
 /* parser_remove_filter */
 int parser_remove_filter(Parser * parser, ParserFilter filter)
-	/* FIXME untested */
+	/* XXX untested */
 {
 	size_t i;
 	ParserFilterData * p;
