@@ -1,17 +1,30 @@
 /* $Id$ */
-/* Copyright (c) 2005-2016 Pierre Pronchery <khorben@defora.org> */
+/* Copyright (c) 2005-2021 Pierre Pronchery <khorben@defora.org> */
 /* This file is part of DeforaOS System libSystem */
-/* This program is free software: you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License as published by
- * the Free Software Foundation, version 3 of the License.
+/* All rights reserved.
  *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are
+ * met:
  *
- * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>. */
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ *
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
+ * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
+ * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 
 
 
@@ -20,6 +33,7 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <limits.h>
 #include <errno.h>
 #include "System/error.h"
 #include "System/object.h"
@@ -35,7 +49,7 @@ String * string_new(String const * string)
 
 	if(string == NULL)
 	{
-		error_set_code(1, "%s", strerror(EINVAL));
+		error_set_code(-EINVAL, "%s", strerror(EINVAL));
 		return NULL;
 	}
 	if(string_set(&ret, string) != 0)
@@ -47,13 +61,24 @@ String * string_new(String const * string)
 /* string_new_append */
 String * string_new_append(String const * string, ...)
 {
-	String * ret = NULL;
+	String * ret;
 	va_list ap;
 
+	va_start(ap, string);
+	ret = string_new_appendv(string, ap);
+	va_end(ap);
+	return ret;
+}
+
+
+/* string_new_appendv */
+String * string_new_appendv(String const * string, va_list ap)
+{
+	String * ret = NULL;
+
 	if(string == NULL)
 		return string_new("");
 	ret = string_new(string);
-	va_start(ap, string);
 	for(string = va_arg(ap, String *); string != NULL;
 			string = va_arg(ap, String *))
 		if(string_append(&ret, string) != 0)
@@ -62,7 +87,6 @@ String * string_new_append(String const * string, ...)
 			ret = NULL;
 			break;
 		}
-	va_end(ap);
 	return ret;
 }
 
@@ -72,28 +96,44 @@ String * string_new_format(String const * format, ...)
 {
 	String * ret;
 	va_list ap;
+
+	va_start(ap, format);
+	ret = string_new_formatv(format, ap);
+	va_end(ap);
+	return ret;
+}
+
+
+/* string_new_formatv */
+String * string_new_formatv(String const * format, va_list ap)
+{
+	String * ret;
+	va_list v;
 	int len;
 	size_t s;
 
-	va_start(ap, format);
-	len = vsnprintf(NULL, 0, format, ap);
-	va_end(ap);
+	if(format == NULL)
+	{
+		error_set_code(-EINVAL, "%s", strerror(EINVAL));
+		return NULL;
+	}
+	va_copy(v, ap);
+	len = vsnprintf(NULL, 0, format, v);
+	va_end(v);
 	if(len < 0)
 	{
-		error_set_code(1, "%s", strerror(errno));
+		error_set_code(-errno, "%s", strerror(errno));
 		return NULL;
 	}
 	s = (size_t)len + 1;
-	if((ret = object_new(s)) == NULL)
+	if((ret = (String *)object_new(s)) == NULL)
 		return NULL;
-	va_start(ap, format);
 	if(vsnprintf(ret, s, format, ap) != len)
 	{
-		error_set_code(1, "%s", strerror(errno));
+		error_set_code(-errno, "%s", strerror(errno));
 		object_delete(ret);
 		ret = NULL;
 	}
-	va_end(ap);
 	return ret;
 }
 
@@ -106,9 +146,14 @@ String * string_new_length(String const * string, size_t length)
 #ifdef DEBUG
 	fprintf(stderr, "DEBUG: %s(\"%s\", %zu)\n", __func__, string, length);
 #endif
-	if((ret = object_new(++length)) == NULL)
+	if(length + 1 == 0)
+	{
+		error_set_code(-ERANGE, "%s", strerror(ERANGE));
 		return NULL;
-	snprintf(ret, length, "%s", string);
+	}
+	if((ret = (String *)object_new(length + 1)) == NULL)
+		return NULL;
+	snprintf(ret, length + 1, "%s", (string != NULL) ? string : "");
 	return ret;
 }
 
@@ -138,6 +183,15 @@ void string_delete(String * string)
 
 
 /* accessors */
+/* string_get */
+# undef string_get
+String const * string_get(String const * string)
+{
+	return string;
+}
+# define string_get(a) (a)
+
+
 /* string_get_length */
 size_t string_get_length(String const * string)
 {
@@ -160,10 +214,9 @@ int string_set(String ** string, String const * string2)
 {
 	size_t len = string_get_length(string2);
 
-	if(object_resize((Object**)string, len + 1) != 0)
+	if(object_resize((Object **)string, len + 1) != 0)
 		return 1;
-	strncpy(*string, string2, len);
-	(*string)[len] = '\0';
+	strcpy(*string, string2);
 	return 0;
 }
 
@@ -176,7 +229,7 @@ int string_append(String ** string, String const * append)
 	size_t alength;
 
 	if(append == NULL)
-		return error_set_code(1, "%s", strerror(EINVAL));
+		return error_set_code(-EINVAL, "%s", strerror(EINVAL));
 	if((alength = string_get_length(append)) == 0)
 		return 0;
 	if(object_resize((Object**)string, slength + alength + 1) != 0)
@@ -186,6 +239,47 @@ int string_append(String ** string, String const * append)
 }
 
 
+/* string_append_format */
+int string_append_format(String ** string, String const * format, ...)
+{
+	int ret;
+	va_list ap;
+
+	va_start(ap, format);
+	ret = string_append_formatv(string, format, ap);
+	va_end(ap);
+	return ret;
+}
+
+
+/* string_append_formatv */
+int string_append_formatv(String ** string, String const * format, va_list ap)
+{
+	va_list v;
+	int alen;
+	size_t len;
+	size_t s;
+
+	if(format == NULL)
+		return error_set_code(-EINVAL, "%s", strerror(EINVAL));
+	va_copy(v, ap);
+	alen = vsnprintf(NULL, 0, format, v);
+	va_end(v);
+	if(alen < 0)
+		return error_set_code(-errno, "%s", strerror(errno));
+	len = string_get_length(*string);
+	s = (size_t)alen + 1;
+	if(object_resize((Object **)string, len + s) != 0)
+		return -1;
+	if(vsnprintf(&(*string)[len], s, format, ap) != alen)
+	{
+		error_set_code(-errno, "%s", strerror(errno));
+		return -1;
+	}
+	return 0;
+}
+
+
 /* string_clear */
 void string_clear(String * string)
 {
@@ -241,12 +335,12 @@ int string_compare_length(String const * string, String const * string2,
 
 
 /* string_explode */
-/* FIXME return a StringArray instead? */
-String ** string_explode(String const * string, String const * separator)
+static void _explose_foreach_delete(ArrayData * value, void * data);
+
+StringArray * string_explode(String const * string, String const * separator)
 {
-	String ** ret = NULL;
-	size_t ret_cnt = 0;
-	String ** p;			/* temporary pointer */
+	StringArray * ret;
+	String * p;			/* temporary pointer */
 	size_t i;			/* current position */
 	String const * s;		/* &string[i] */
 	ssize_t j;			/* position of the next separator */
@@ -256,9 +350,12 @@ String ** string_explode(String const * string, String const * separator)
 	fprintf(stderr, "DEBUG: %s(\"%s\", \"%s\")\n", __func__, string,
 			separator);
 #endif
+	if((ret = stringarray_new()) == NULL)
+		return NULL;
 	if(separator == NULL || (l = string_get_length(separator)) == 0)
 	{
-		error_set_code(1, "%s", strerror(EINVAL));
+		error_set_code(-EINVAL, "%s", strerror(EINVAL));
+		array_delete(ret);
 		return NULL;
 	}
 	for(i = 0;; i += j + l)
@@ -268,34 +365,45 @@ String ** string_explode(String const * string, String const * separator)
 #ifdef DEBUG
 		fprintf(stderr, "DEBUG: %s(): i=%zu, j=%zd\n", __func__, i, j);
 #endif
-		if((p = realloc(ret, sizeof(*ret) * (ret_cnt + 2))) == NULL)
-			break;
-		ret = p;
 		if(j < 0)
 		{
-			if((ret[ret_cnt++] = string_new(s)) == NULL)
+			if((p = string_new(s)) == NULL
+					|| array_append(ret, p) != 0)
+			{
+				string_delete(p);
 				break;
+			}
 #ifdef DEBUG
 			fprintf(stderr, "DEBUG: %s(): \"%s\"\n", __func__,
 					ret[ret_cnt - 1]);
 #endif
-			ret[ret_cnt++] = NULL;
 			return ret;
 		}
-		if((ret[ret_cnt++] = string_new_length(s, j)) == NULL)
+		if((p = string_new_length(s, j)) == NULL
+				|| array_append(ret, p) != 0)
+		{
+			string_delete(p);
 			break;
+		}
 #ifdef DEBUG
 		fprintf(stderr, "DEBUG: %s(): \"%s\"\n", __func__,
 				ret[ret_cnt - 1]);
 #endif
 	}
 	/* free everything */
-	for(p = ret; *p != NULL; p++)
-		string_delete(*p);
-	free(ret);
+	array_foreach(ret, _explose_foreach_delete, NULL);
+	array_delete(ret);
 	return NULL;
 }
 
+static void _explose_foreach_delete(ArrayData * value, void * data)
+{
+	String * s = (String *)value;
+	(void) data;
+
+	string_delete(s);
+}
+
 
 /* string_find */
 String * string_find(String const * string, String const * key)
@@ -312,15 +420,18 @@ String * string_find(String const * string, String const * key)
 ssize_t string_index(String const * string, String const * key)
 {
 	size_t len;
-	ssize_t i;
+	size_t keylen;
+	size_t i;
 
-	len = string_get_length(key);
-	for(i = 0; string[i] != '\0'
-			&& string_compare_length(&string[i], key, len) != 0;
-			i++);
-	if(string[i] == '\0')
+	len = string_get_length(string);
+	if((keylen = string_get_length(key)) == 0)
+		return len;
+	if(keylen > len)
 		return -1;
-	return i;
+	for(i = 0; i <= len - keylen; i++)
+		if(string_compare_length(&string[i], key, keylen) == 0)
+			return i;
+	return -1;
 }
 
 
@@ -385,6 +496,25 @@ int string_replace(String ** string, String const * what, String const * by)
 }
 
 
+/* string_rindex */
+ssize_t string_rindex(String const * string, String const * key)
+{
+	size_t len;
+	size_t keylen;
+	ssize_t i;
+
+	len = string_get_length(string);
+	if((keylen = string_get_length(key)) == 0)
+		return len;
+	if(keylen > len)
+		return -1;
+	for(i = len - keylen; i >= 0; i--)
+		if(string_compare_length(&string[i], key, keylen) == 0)
+			return i;
+	return -1;
+}
+
+
 /* string_rtrim */
 size_t string_rtrim(String * string, String const * which)
 {
@@ -416,6 +546,26 @@ size_t string_rtrim(String * string, String const * which)
 }
 
 
+/* string_tolower */
+void string_tolower(String * string)
+{
+	size_t i;
+
+	for(i = string_get_length(string); i > 0; i--)
+		string[i - 1] = tolower((unsigned char)string[i - 1]);
+}
+
+
+/* string_toupper */
+void string_toupper(String * string)
+{
+	size_t i;
+
+	for(i = string_get_length(string); i > 0; i--)
+		string[i - 1] = toupper((unsigned char)string[i - 1]);
+}
+
+
 /* string_trim */
 size_t string_trim(String * string, String const * which)
 {
